//********************************************************************
//  BookOfWar.java @version 1.20
//    Simulate Book of War battles for cost-balancing purposes.
//  Copyright (c) 2009-11 Daniel R. Collins. All rights reserved.
//  See the bottom of this file for any licensing information.
//********************************************************************

import java.awt.Toolkit;

public class BookOfWar {

	//-----------------------------------------------------------------
	//  Combat properties
	//-----------------------------------------------------------------
	int distance;
	Weather weather;
	Terrain terrain;
	boolean pikeFlanking;
	boolean pikeInterrupted;
	Unit winner;

	enum Weather {Sunny, Cloudy, Rainy, Stormy};
	enum Terrain {Gulley, Rough, Hill, Open, Woods, Marsh, Stream, Pond}

	//-----------------------------------------------------------------
	//  Constant parameters
	//-----------------------------------------------------------------
	final int budgetMin = 50; // Base 50
	final int budgetMax = 200; // Base 200

	// Switch this to increase assessment accuracy (slow)
	final boolean useCloseAssessment = false;

	// Switch this to maximize assessment accuracy (very slow)
	final boolean useExactAssessment = false;

	// Switch this to see one game played out in detail
	static final boolean gameDetail = false;

	// Other stuff we always assume turned on
	final boolean useRangePenalty = true;
	final boolean useSpecialAbilities = true;
	final boolean useOptionalMoraleMods = true;

	// This alters core pike effectiveness vs. swords & cavalry, etc.
	final double pikeFlankingChance = 0.30;

	// This alters core sword effectiveness vs. pike & cavalry, etc.
	final double terrainMultiplier = 1.00;

	//-----------------------------------------------------------------
	//  Create table of assessed win percents
	//    Change what kind of units to compare here
	//-----------------------------------------------------------------
	void assessmentTable () {
 		Unit[] array1 = basicUnit;
 		Unit[] array2 = basicUnit;
 		makeAssessmentTable(array1, array2);
	}

	//-----------------------------------------------------------------
	//  Battle two specified unit types with detail reports
	//    Change what units to battle here
	//-----------------------------------------------------------------
	void gameDetail () {
		Unit unit1 = coreUnit[0];
		Unit unit2 = coreUnit[1];
		oneGame(unit1, unit2);
	}

	//-----------------------------------------------------------------
	//  Basic Unit types available
	//-----------------------------------------------------------------
	Unit[] coreUnit = {
 		new Unit("Pikes", Unit.Type.infantry, 5, 12, 4, 1),
 		new Unit("Swords", Unit.Type.infantry, 5, 9, 5, 1),
 		new Unit("Cavalry", Unit.Type.cavalry, 18, 12, 6, 2),
	};

	Unit[] coreUnit2 = {
 		new Unit("Pikes", Unit.Type.infantry, 5, 12, 4, 1),
 		new Unit("Swords", Unit.Type.infantry, 5, 9, 5, 1),
 		new Unit("Cavalry", Unit.Type.cavalry, 18, 12, 6, 2),
		new Unit("Longbows", Unit.Type.archers, 10, 12, 4, 1, 2, 21), 
		new Unit("Horse Archers", Unit.Type.cavalry, 13, 24, 4, 2, 2, 18), 
	};

	Unit[] basicUnit = {

		// Infantry
		new Unit("Light Infantry", Unit.Type.infantry, 4, 12, 4, 1), 
		new Unit("Medium Infantry", Unit.Type.infantry, 5, 9, 5, 1), 
		new Unit("Heavy Infantry", Unit.Type.infantry, 9, 6, 6, 1),
		new Unit("Pikemen", Unit.Type.infantry, 5, 12, 4, 1),	

		// Archers
		new Unit("Archers", Unit.Type.archers, 7, 12, 4, 1, 2, 15),
		new Unit("Longbowmen", Unit.Type.archers, 10, 12, 4, 1, 2, 21), 
		new Unit("Crossbowmen", Unit.Type.archers, 5, 12, 4, 1, 1, 18), 
		new Unit("Heavy Crossbows", Unit.Type.archers, 8, 9, 5, 1, 1, 24),

		// Cavalry	
		new Unit("Light Cavalry", Unit.Type.cavalry, 9, 24, 4, 2),
		new Unit("Medium Cavalry", Unit.Type.cavalry, 12, 18, 5, 2),
		new Unit("Heavy Cavalry", Unit.Type.cavalry, 18, 12, 6, 2),	
		new Unit("Horse Archers", Unit.Type.cavalry, 13, 24, 4, 2, 2, 18), 
	};

	//-----------------------------------------------------------------
	//  Fantasy Unit types available
	//-----------------------------------------------------------------
	Unit[] fantasyUnit = {

		// Lawful Units: Basic
		new Unit("Halfling Lt Infantry", Unit.Type.infantry, Unit.Alignment.Lawful, 0.75, 
			4, 12, 4, 1,  0, 0),
		new Unit("Halfling Slingers", Unit.Type.archers, Unit.Alignment.Lawful, 0.75,
			6, 12, 4, 1, 1, 12),
		new Unit("Elf Lt Infantry", Unit.Type.infantry, Unit.Alignment.Lawful, 0.75,
			4, 12, 4, 1, 0, 0),
		new Unit("Elf Med Infantry", Unit.Type.infantry, Unit.Alignment.Lawful, 0.75,
			5, 9, 5, 1, 0, 0),
		new Unit("Elf Archers", Unit.Type.archers, Unit.Alignment.Lawful, 0.75,
			7, 12, 4, 1, 2, 15),
		new Unit("Dwarf Med Infantry", Unit.Type.infantry, Unit.Alignment.Lawful, 0.75,
			5, 9, 5, 1, 0, 0),
		new Unit("Dwarf Hvy Crossbows", Unit.Type.archers, Unit.Alignment.Lawful, 0.75,
			8, 9, 5, 1, 1, 24),
		new Unit("Dwarf Cranequiniers", Unit.Type.cavalry, Unit.Alignment.Lawful, 1.00,
			13, 12, 5, 2, 1, 18),

		// Chaotic Units: Basic
		new Unit("Goblin Lt Infantry", Unit.Type.infantry, Unit.Alignment.Chaotic, 0.75, 
			3, 9, 4, 1, 0, 0),
		new Unit("Goblin Archers", Unit.Type.archers, Unit.Alignment.Chaotic, 0.75, 
			6, 9, 4, 1, 2, 15), 
		new Unit("Goblin Wolf Riders", Unit.Type.cavalry,  Unit.Alignment.Chaotic, 1.00, 
			10, 12, 4, 3, 0, 0), 
		new Unit("Orc Lt Infantry", Unit.Type.infantry, Unit.Alignment.Chaotic, 0.75, 
			3, 12, 4, 1, 0, 0), 
		new Unit("Orc Med Infantry", Unit.Type.infantry, Unit.Alignment.Chaotic, 0.75, 
			4, 9, 5, 1, 0, 0), 
		new Unit("Orc Archers", Unit.Type.archers, Unit.Alignment.Chaotic, 0.75, 
			6, 12, 4, 1, 2, 15), 
		new Unit("Gnoll Med Infantry", Unit.Type.infantry, Unit.Alignment.Chaotic, 1.00, 
			8, 9, 5, 2, 0, 0), 
		new Unit("Gnoll Longbowmen", Unit.Type.archers, Unit.Alignment.Chaotic, 1.00, 
			14, 9, 5, 2, 2, 21), 

		// Lawful Units: Elite
		new Unit("Halfling Elite Slingers", Unit.Type.archers, Unit.Alignment.Lawful, 0.75,
			16, 12, 4, 3, 1, 12), // Slings +1, hide in woods
		new Unit("Elf Elite Archers", Unit.Type.archers, Unit.Alignment.Lawful, 0.75,
			30, 12, 5, 3, 2, 21), // Elfin chain, invisibility
		new Unit("Dwarf Elite Infantry", Unit.Type.infantry, Unit.Alignment.Lawful, 0.75, 
			40, 9, 6, 3,  0, 0), // Dwarven plate, dodge giants 
			// Ideal price 40 vs. basic, 35 vs. fantasy
		new Unit("Men Elite Cavalry", Unit.Type.cavalry, Unit.Alignment.Lawful, 1.00,
			90, 12, 6, 6, 0, 0), // 2 attacks (1d6+1 each) 
			// Ideal price 90 vs. basic, 80 vs. fantasy

		// Chaotic Units: Elite
		new Unit("Bugbears", Unit.Type.infantry, Unit.Alignment.Chaotic, 1.25, 
			12, 9, 5, 3, 0, 0),
		new Unit("Ogres", Unit.Type.infantry, Unit.Alignment.Chaotic, 1.50, 
			18, 9, 5, 4, 0, 0),  // Great strength (damage 1.5) 
			// Ideal price 18 vs. basic, 16 vs. fantasy
		new Unit("Trolls", Unit.Type.infantry, Unit.Alignment.Chaotic, 1.50,
			75, 12, 5, 6, 0, 0), // Fearless, regenerate, 2 atks
		new Unit("Giant, Hill", Unit.Type.archers, Unit.Alignment.Chaotic, 1.50,
			55, 12, 5, 8, 1, 20), // Throw rocks, damage 2
	};

	//-----------------------------------------------------------------
	//  Hero Unit types available
	//-----------------------------------------------------------------
	Hero[] heroUnit = {

		new Hero("Knight Renown",        Unit.Type.cavalry,  40, 12, 7, 1, 1, 2), 
		new Hero("Knight Commander",     Unit.Type.cavalry,  60, 12, 7, 2, 1, 3), 
		new Hero("Barbarian Lord",       Unit.Type.infantry, 30, 12, 6, 1, 3, 3),
		new Hero("Barbarian King",       Unit.Type.infantry, 55, 12, 7, 2, 3, 3), 
		new Hero("Dragon, Adult Blue",   Unit.Type.infantry, 40, 24, 6, 1, 3, 2),
		new Hero("Dragon, Adult Red",    Unit.Type.infantry, 40, 24, 6, 1, 3, 2), 
		new Hero("Dragon, Adult Gold",   Unit.Type.infantry, 40, 24, 6, 1, 3, 2), 
		new Hero("Giant, Frost",         Unit.Type.infantry, 6, 12, 5, 1, 5, 2),
		new Hero("Giant, Fire",          Unit.Type.infantry, 6, 12, 5, 1, 5, 3), 
		new Hero("Giant, Storm",         Unit.Type.infantry, 7, 15, 5, 1, 5, 4), 

		/*
			Note: Hero missile weapons are ignored.
			Pricing heroes requires a number of judgement calls.
			Embedded usage can be significantly more powerful
			  (esp. Bar. Lord, weak solo, doubled cost for embedded strength).
			Other modifiers applied to book prices (e.g., Storm Giant weather control)
		*/
	};

	//-----------------------------------------------------------------
	//  Basic Units with attached Hero
	//-----------------------------------------------------------------
	Unit[] heroLedUnit = {
 		new Unit("Pikes Hero 1", Unit.Type.infantry, 5, 12, 4, 1, heroUnit[0]),
 		new Unit("Swords Hero 1", Unit.Type.infantry, 5, 9, 5, 1, heroUnit[0]),
 		new Unit("Cavalry Hero 1", Unit.Type.cavalry, 18, 12, 6, 2, heroUnit[0]),
		
 		new Unit("Pikes Hero 2", Unit.Type.infantry, 5, 12, 4, 1, heroUnit[1]),
 		new Unit("Swords Hero 2", Unit.Type.infantry, 5, 9, 5, 1, heroUnit[1]),
 		new Unit("Cavalry Hero 2", Unit.Type.cavalry, 18, 12, 6, 2, heroUnit[1]),

 		new Unit("Pikes Hero 3", Unit.Type.infantry, 5, 12, 4, 1, heroUnit[2]),
 		new Unit("Swords Hero 3", Unit.Type.infantry, 5, 9, 5, 1, heroUnit[2]),
 		new Unit("Cavalry Hero 3", Unit.Type.cavalry, 18, 12, 6, 2, heroUnit[2]),

 		new Unit("Pikes Hero 4", Unit.Type.infantry, 5, 12, 4, 1, heroUnit[3]),
 		new Unit("Swords Hero 4", Unit.Type.infantry, 5, 9, 5, 1, heroUnit[3]),
 		new Unit("Cavalry Hero 4", Unit.Type.cavalry, 18, 12, 6, 2, heroUnit[3]),
	};

	//-----------------------------------------------------------------
	//  Wizard Unit types available
	//-----------------------------------------------------------------
	Hero[] wizardUnit = {
		new Hero("Wizard Rank 1", Unit.Type.infantry, 80, 12, 7, 1, 6, 1, 1), 
 		new Hero("Wizard Rank 2", Unit.Type.infantry, 90, 12, 7, 1, 6, 1, 2), 
 		new Hero("Wizard Rank 3", Unit.Type.infantry, 100, 12, 7, 1, 6, 1, 3), 
 		new Hero("Wizard Rank 4", Unit.Type.infantry, 110, 12, 7, 1, 6, 1, 4), 
 		new Hero("Wizard Rank 5", Unit.Type.infantry, 120, 12, 7, 1, 6, 1, 5), 
 		new Hero("Wizard Rank 6", Unit.Type.infantry, 130, 12, 7, 1, 6, 1, 6), 
	};

	//-----------------------------------------------------------------
	//  Basic Units with attached Wizard
	//-----------------------------------------------------------------
	Unit[] wizardLedUnit = {
  		new Unit("Pikes Wizard 1", Unit.Type.infantry, 5, 12, 4, 1, wizardUnit[0]),
  		new Unit("Pikes Wizard 2", Unit.Type.infantry, 5, 12, 4, 1, wizardUnit[1]),
  		new Unit("Pikes Wizard 3", Unit.Type.infantry, 5, 12, 4, 1, wizardUnit[2]),
  		new Unit("Pikes Wizard 4", Unit.Type.infantry, 5, 12, 4, 1, wizardUnit[3]),
  		new Unit("Pikes Wizard 5", Unit.Type.infantry, 5, 12, 4, 1, wizardUnit[4]),
  		new Unit("Pikes Wizard 6", Unit.Type.infantry, 5, 12, 4, 1, wizardUnit[5]),

 		new Unit("Swords Wizard 1", Unit.Type.infantry, 5, 9, 5, 1, wizardUnit[0]),
 		new Unit("Swords Wizard 2", Unit.Type.infantry, 5, 9, 5, 1, wizardUnit[1]),
 		new Unit("Swords Wizard 3", Unit.Type.infantry, 5, 9, 5, 1, wizardUnit[2]),
 		new Unit("Swords Wizard 4", Unit.Type.infantry, 5, 9, 5, 1, wizardUnit[3]),
 		new Unit("Swords Wizard 5", Unit.Type.infantry, 5, 9, 5, 1, wizardUnit[4]),
 		new Unit("Swords Wizard 6", Unit.Type.infantry, 5, 9, 5, 1, wizardUnit[5]),

  		new Unit("Cavalry Wizard 1", Unit.Type.cavalry, 18, 12, 6, 2, wizardUnit[0]),
  		new Unit("Cavalry Wizard 2", Unit.Type.cavalry, 18, 12, 6, 2, wizardUnit[1]),
  		new Unit("Cavalry Wizard 3", Unit.Type.cavalry, 18, 12, 6, 2, wizardUnit[2]),
  		new Unit("Cavalry Wizard 4", Unit.Type.cavalry, 18, 12, 6, 2, wizardUnit[3]),
  		new Unit("Cavalry Wizard 5", Unit.Type.cavalry, 18, 12, 6, 2, wizardUnit[4]),
  		new Unit("Cavalry Wizard 6", Unit.Type.cavalry, 18, 12, 6, 2, wizardUnit[5]),
	};

	//-----------------------------------------------------------------
	//  Report detail for detailed game
	//-----------------------------------------------------------------
	void reportDetail (String s) {
		if (gameDetail) 
			System.out.println(s);
	}
	
	//-----------------------------------------------------------------
	//  Printf recreation (for copied tables code)
	//-----------------------------------------------------------------
	void printf(String s) {
		System.out.print(s);	
	}

	//-----------------------------------------------------------------
	//  Make general assessment table
	//-----------------------------------------------------------------
	void makeAssessmentTable (Unit[] array1, Unit[] array2) {
  
		// Title
		printf("Assessed win percents "
			+ "(budget " + budgetMin + "-" + budgetMax + "):\n\n");

	 	// Header
		for (int i = 0; i < array1.length; i++) {
			printf("\t" + array1[i].getAbbreviation());
		}
		printf("\tWins\n");
		for (int i = 0; i < array1.length + 1; i++) {
			printf("--------");
		}
		printf("\n");

  		// Body
		for (int i = 0; i < array2.length; i++) {
			int winCount = 0;
			printf(array2[i].getAbbreviation() + "\t");
			for (int j = 0; j < array1.length; j++) {
				if (array1[j] == array2[i])
					printf("-\t");
				else {
					int pctWon = (int) 
						(masterAssessGames(array2[i], array1[j]) * 100 + 0.5);
					printf(pctWon >= 50 ? pctWon + "\t" : "-\t");
					if (pctWon >= 50) winCount++;
				}
			}
			printf(winCount + "\n");
		}

		// Tail
		printf("\n");
	}

	//-----------------------------------------------------------------
	//  Set ranks and files
	//-----------------------------------------------------------------
	void setRanksAndFiles (Unit unit) {

		// Pick optimal row size
		int ranks = 1;
		switch (unit.getType()) {
			case infantry: ranks = 4; break;
			case archers: ranks = 1; break;
			case cavalry: ranks = 1; break;
		}

		// Force heroes to 1 rank
		if (unit instanceof Hero) ranks = 1;

		// Force front at least as wide
		int sqrt = (int)(Math.sqrt(unit.getFigures()));
		if (sqrt < ranks) ranks = Math.max(sqrt, 1);
		unit.setFilesByRanks(ranks);
	}

	//-----------------------------------------------------------------
	//  Initialize one unit by budget
	//-----------------------------------------------------------------
	void initUnit (Unit unit, int budget) {

		// Buy any attached hero
		if (unit.hasHero()) {
			Hero hero = unit.getHero();
			if (budget >= hero.getCost()) {
				initUnit(hero, hero.getCost());
				budget -= hero.getCost();
			}
			else {
				initUnit(hero, 0);			
			}
		}

		// Buy & set up normal figures
		unit.setFigures(budget/unit.getCost());
		setRanksAndFiles(unit);

		// Prepare any hero abilities
		if (unit instanceof Hero)
			((Hero) unit).refreshCharges();
	}

	//-----------------------------------------------------------------
	//  Randomize budget & initialize opposing units
	//-----------------------------------------------------------------
	void initUnitsByBudget (Unit unit1, Unit unit2) {
		int range = budgetMax - budgetMin;
		int budget = budgetMin + (int) (Math.random() * range);
		reportDetail("Budget: " + budget);
		initUnit(unit1, budget);
		initUnit(unit2, budget);
	}

	//-----------------------------------------------------------------
	//  Initialize battlefield (terrain, weather, distance, etc.)
	//-----------------------------------------------------------------
	void initBattlefield () {
		randomizeTerrain();
		randomizeWeather();
		distance = 25 + (int) (Math.random() * 25);
		pikeFlanking = Math.random() < pikeFlankingChance;
		reportDetail("Terrain: " + terrain);
		reportDetail("Weather: " + weather);
		reportDetail("Distance: " + distance);
		reportDetail("Flanking: " + pikeFlanking);
		pikeInterrupted = false;
		winner = null;
	}

	//-----------------------------------------------------------------
	//  Run many game trials; report rate of wins by first unit
	//-----------------------------------------------------------------
	double assessGames(Unit unit1, Unit unit2, int trials) {
		int wins = 0;
		for (int i = 0; i < trials; i++) {
			oneGame(unit1, unit2);
			if (unit1 == winner) wins++;
		}
		return (double) wins/trials;
	}

	//-----------------------------------------------------------------
	//  Find rate of wins by first unit.
	//    Use some statistics here to focus effort when close to 50%.
	//    See Weiss 8E p. 615, "sample size for estimating p". CL 99%.
	//-----------------------------------------------------------------
	double masterAssessGames(Unit unit1, Unit unit2) {
		double p;  // Sample proportion p-hat

		if (useExactAssessment) {
			p = assessGames(unit1, unit2, 16800);  // E < 1%
		}
		else {
			p = assessGames(unit1, unit2, 50);     // E < 20%
				if (p < 0.3 || p > 0.7) return p;
			p = assessGames(unit1, unit2, 200);    // E < 10%
				if (p < 0.4 || p > 0.6) return p;
			p = assessGames(unit1, unit2, 800);    // E < 5%
				if (p < 0.45 || p > 0.55) return p;
			if (useCloseAssessment) {
				p = assessGames(unit1, unit2, 4200);   // E < 2%
					if (p < 0.48 || p > 0.52) return p;
				p = assessGames(unit1, unit2, 16800);  // E < 1%
			}
		}
		return p;
	}

	//-----------------------------------------------------------------
	//  One game
	//-----------------------------------------------------------------
	void oneGame (Unit unit1, Unit unit2) {

		// Set up game
		initBattlefield();
		initUnitsByBudget(unit1, unit2);

		// Initiative for unit2 to start
		if (d6() > 3)
			oneTurn(unit2, unit1);
		
		// Battle until one side wins
		while (winner == null) {
			oneTurn(unit1, unit2);
			if (winner != null) break;
			oneTurn(unit2, unit1);
		}

		// Report on winner
		reportDetail("* WINNER *: " + winner);
	}

	//-----------------------------------------------------------------
	//  One turn of action
	//-----------------------------------------------------------------
	void oneTurn (Unit unit1, Unit unit2) {

		// Initialize
		unit2.clearFigsLostInTurn();

		// Take action by type
		if (unit1.hasAnyWizard())
			oneTurnWizard(unit1, unit2);
		else if (unit1.hasMissiles())
			oneTurnRanged(unit1, unit2);
		else
			oneTurnMelee(unit1, unit2);

		// Check morale
		checkMorale(unit2);

		// Check regeneration
		if (useSpecialAbilities 
			&& unit2.nameStarts("Troll"))
				unit2.regenerate();
			
		// Check for victory
		if (unit1.isBeaten()) winner = unit2;
		if (unit2.isBeaten()) winner = unit1;
	}

	//-----------------------------------------------------------------
	//  Move unit forward given distance
	//-----------------------------------------------------------------
	void moveForward (Unit unit, int move) {
		assert(move > 0);
		distance -= move;
		if (distance < 0) distance = 0;
		reportDetail(unit + " move to dist. " + distance);
	}

	//-----------------------------------------------------------------
	//  Apply damage from attack & report
	//-----------------------------------------------------------------
	void applyDamage (Unit attacker, Unit defender, boolean ranged, int hits) {
		String startReport = attacker 
			+ (ranged ? " shoot " : " attack ") + defender + ": ";

		int figsKilled = defender.takeDamage(hits);

		String endReport = figsKilled + " fig" 
			+ (figsKilled == 1 ? "" : "s") + " killed";
		reportDetail(startReport + endReport);
	}

	//-----------------------------------------------------------------
	//  One turn for melee troops
	//-----------------------------------------------------------------
	void oneTurnMelee (Unit unit1, Unit unit2) {
	
		// Move towards contact
		if (distance > meleeRange(unit1)) {
			int move = getMove(unit1);
			int gap = distance - meleeRange(unit1);
			moveForward(unit1, Math.min(move, gap));
		}

		// Check enemy pike interrupt attack
		if (onPikeInterrupt(unit2)) {
			unit1.clearFigsLostInTurn();
			meleeAttack(unit2, unit1);
			checkMorale(unit1);
			pikeInterrupted = true;
			if (unit1.isBeaten()) return;
		}

		// Attack if in contact
		if (distance <= meleeRange(unit1)) {
			meleeAttack(unit1, unit2);
		}
	}

	//-----------------------------------------------------------------
	//  One turn for ranged troops (AI-dependent)
	//-----------------------------------------------------------------
	void oneTurnRanged (Unit unit1, Unit unit2) {
		int minDist = minDistanceToShoot(unit1, unit2);

		// If shooting impossible or in contact, melee attack
		if (minDist == 0
				|| distance <= meleeRange(unit1)
				|| distance <= meleeRange(unit2)) {
			oneTurnMelee(unit1, unit2);
		}
		// If enemy has no missiles, stay motionless & just shoot
		else if (!unit2.hasMissiles() && !unit2.hasAnyWizard()) {
			if (distance <= minDist)
				rangedAttack(unit1, unit2, distance, true);
		}
		// Possible charge to melee to reduce enemy missile fire
		else if (attackAdvantageRatio(unit1, unit2) > 1) {
			oneTurnMelee(unit1, unit2);
		}
		// If in range, full attack
		else if (distance <= minDist) {
			rangedAttack(unit1, unit2, distance, true);
		}
		// If enemy outranges us, get in range asap
		else if (minDist < minDistanceToShoot(unit2, unit1)
				|| (unit2.hasAnyWizard() && minDist < 24)) {
			int gap = distance - minDist;
			moveForward(unit1, Math.min(getMove(unit1), gap));
			if (distance <= minDist && gap <= getMove(unit1)/2) 
				rangedAttack(unit1, unit2, distance, false);
		}
		// Else half-move to range if possible
		else if (getMove(unit1)/2 > 0) {
			int gap = distance - minDist;
			moveForward(unit1, Math.min(getMove(unit1)/2, gap));
			if (distance <= minDist)
				rangedAttack(unit1, unit2, distance, false);
		}
		// Else can only move 1"
		else {
			assert(getMove(unit1) == 1);
			moveForward(unit1, 1);
		}		
	}

	//-----------------------------------------------------------------
	//  One turn with wizard
	//    No move so wizard can shoot full; ranged or melee as capable.
	//    Note: Currently cannot handle wizard-vs-wizard engagements
	//-----------------------------------------------------------------
	void oneTurnWizard (Unit unit1, Unit unit2) {

		// All-wizard unit
		if (unit1.isWizard()) {
			wizardSpellAttack((Hero)unit1, unit2);
		}

		// Troops with attached wizard
		else {
			if (unit1.hasMissiles() && distance > 0) {
				rangedAttack(unit1, unit2, distance, true);
			}
			else if (distance <= meleeRange(unit1)) {
				meleeAttack(unit1, unit2);
			}	
			oneTurnWizard(unit1.getHero(), unit2);
		}
	}

	//-----------------------------------------------------------------
	//  Range for melee attacks (zero except for pikes)
	//-----------------------------------------------------------------
	int meleeRange (Unit unit) {
		return (unit.hasPikes() ? 1 : 0); 
	}

	//-----------------------------------------------------------------
	//  Attack advantage ratio
	//    Like an odds ratio for melee/range attack rate.
	//    Over 1 prefer melee; under 1 prefer ranged.
	//    Both units must have missiles (else divide by zero)
	//-----------------------------------------------------------------
	double attackAdvantageRatio (Unit unit1, Unit unit2) {
		int rateOfMelee1 = meleeAttackDice(unit1, unit2, 1);
		int rateOfMelee2 = meleeAttackDice(unit2, unit1, 1);
		double atkAdvantage1 = (double) rateOfMelee1 / unit1.getROF();
		double atkAdvantage2 = (double) rateOfMelee2 / unit2.getROF();
		return atkAdvantage1 / atkAdvantage2;
	}

	//-----------------------------------------------------------------
	//  Minimum distance to have chance of shooting enemy
	//-----------------------------------------------------------------
	int minDistanceToShoot (Unit unit1, Unit unit2) {
		int baseToHit = unit2.getAH() - unit1.getHD()/3 
			- miscAtkBonus(unit1, unit2, true);
		if (baseToHit > 6 || !terrainPermitShots())
			return 0;                     // Melee only
		else if (baseToHit == 6 && useRangePenalty)
			return unit1.getRA() / 2;     // Short only
		else
			return unit1.getRA();         // Full range
	}

	//-----------------------------------------------------------------
	//  Does terrain and weather permit shooting?
	//-----------------------------------------------------------------
	boolean terrainPermitShots () {
		return (!(weather == Weather.Stormy
			|| terrain == Terrain.Woods || terrain == Terrain.Gulley));	
	}

	//-----------------------------------------------------------------
	//  Maximum distance to move, including terrain & weather
	//-----------------------------------------------------------------
	int getMove (Unit unit) {
		int moveCost = 1;

		// Terrain mods
		switch (terrain) {
			case Hill: case Gulley: // Assume uphill; +1 step per 2"
			case Rough: case Woods: moveCost *= 2; break;
			case Marsh: moveCost *= 3; break;
			case Stream: moveCost *= 4;
		}
			
		// Weather mods
		if (weather == Weather.Stormy)
			moveCost *= 2;		

		// Cavalry penalties double
		if (moveCost > 1 && unit.getType() == Unit.Type.cavalry)
			moveCost *= 2;

		// Dragons flying ignore all this stuff
		if (useSpecialAbilities && unit.nameStarts("Dragon"))
			moveCost = 1;		

		// Return (at least 1 inch)
		int move = unit.getMV() / moveCost;
		return Math.max(move, 1);
	}

	//-----------------------------------------------------------------
	//  One melee attack; returns number killed
	//-----------------------------------------------------------------
	void meleeAttack (Unit attacker, Unit defender) {

		// Check for heroes attacking
		if (attacker.hasHero()) {
			heroMeleeAttack(attacker.getHero(), defender);
		}
		else if (attacker instanceof Hero) {
			heroMeleeAttack((Hero) attacker, defender);
			return;
		}

		// Check for defender immune
		if (isAttackImmune(attacker, defender))
			return;

		// Compute figures & dice in attack
		double atkWidth = attacker.getFiles() * attacker.getFigWidth();
		double defWidth = defender.getFiles() * defender.getFigWidth();
		int figsAtk = (atkWidth <= defWidth ? attacker.getFiles() :
			(int) Math.ceil(defWidth / attacker.getFigWidth()));
		if ((defender instanceof Hero) && (figsAtk > defender.getFiles()))
			figsAtk = defender.getFiles();
		int atkDice = meleeAttackDice(attacker, defender, figsAtk);

		// Roll attack dice
		int numHits = 0;
		for (int i = 0; i < atkDice; i++) {
			if (rollToHit(defender.getAH(), attacker.getHD(), 
				miscAtkBonus(attacker, defender, false))) numHits++;
		}

		// Extra damage by giant-types
		if (useSpecialAbilities) {
			if (attacker.nameStarts("Ogre") && defender.getHD() > 1)
				numHits = numHits + numHits/2;
			if (attacker.nameStarts("Giant") && defender.getHD() > 1)
				numHits = numHits * 2;
		}

		// Apply damage
		applyDamage(attacker, defender, false, numHits);
	}

	//-----------------------------------------------------------------
	//  One ranged attack; returns number killed
	//-----------------------------------------------------------------
	void rangedAttack (Unit attacker, Unit defender, int range, boolean fullROF) {

		// Check for defender immune
		if (isAttackImmune(attacker, defender))
			return;

		// Compute figures & dice in attack
		int figsAtk = (attacker.getRanks() <= 3 ?
			attacker.getFigures() : attacker.getFiles() * 3);
		int atkDice = figsAtk * attacker.getROF();
		if (!fullROF) atkDice /= 2;

		// Determine range modifier (if any)
		int rangeMod = 0;
		if (range > attacker.getRA())
			return; // out-of-range
		if (useRangePenalty && range > attacker.getRA()/2)
			rangeMod = -1;

		// Roll attack dice
		int numHits = 0;
		for (int i = 0; i < atkDice; i++) {
			if (rollToHit(defender.getAH(), attacker.getHD(), 
				rangeMod + miscAtkBonus(attacker, defender, true))) numHits++;
		}

		// Extra damage by giant-types
		if (useSpecialAbilities) {
			if (attacker.nameStarts("Giant") && defender.getHD() > 1)
				numHits = numHits * 2;
		}

		// Apply damage
		applyDamage(attacker, defender, true, numHits);
	}

	//-----------------------------------------------------------------
	//  Melee attack dice (special modifiers)
	//-----------------------------------------------------------------
	int meleeAttackDice (Unit atk, Unit def, int figsAtk) {

		// Initialize
		int atkDice = figsAtk;

		// Cavalry get double dice in open
		if (atk.getType() == Unit.Type.cavalry
				&& terrain == Terrain.Open 
				&& weather != Weather.Stormy) {
			atkDice *= 2;
		}

		// Pike interrupt attack gets multiplied dice
		if (onPikeInterrupt(atk)) {
			atkDice *= (atk.getRanks() == 1 ? 2 : 4);
		}

		// Creatures with multiple attacks
		if (useSpecialAbilities) {
			if (atk.nameStarts("Troll"))
				atkDice *= 2;
		}

		// Return (at least 1 die)
		return Math.max(atkDice, 1);
	}

	//-----------------------------------------------------------------
	//  Miscellaneous to-hit bonuses
	//-----------------------------------------------------------------
	int miscAtkBonus (Unit unit, Unit enemy, boolean ranged) {
		int bonus = 0;

		// Pike interrupt attack gets +1 infantry, +2 cavalry
		if (onPikeInterrupt(unit)) {
			bonus += (enemy.getType() != Unit.Type.cavalry ? 1 : 2);
		}

		// Solo heroes attacked in normal melee
		if (!ranged && (enemy instanceof Hero) && !(unit instanceof Hero))
			bonus += 1;

		// Rainy weather missile penalty
		if (ranged && weather == Weather.Rainy) 
			bonus -= 1;

		// Racial special abilities
		if (useSpecialAbilities) {

			// Halfling ranged bonus
			if (unit.nameStarts("Halfling") && ranged) bonus += 1;

			// Dwarf dodge giants
			if (enemy.nameStarts("Dwarf") && unit.getHD() >= 4
					&& unit.getAlignment() == Unit.Alignment.Chaotic)
				bonus -= 1;

			// Cavalry inflated HD fix
			if (unit.nameStarts("Goblin Wolf Riders")) bonus -= 1;
			if (unit.nameStarts("Men Elite Cavalry")) bonus -= 1;
		}

		return bonus;
	}
	
	//-----------------------------------------------------------------
	//  Hero melee attack
	//-----------------------------------------------------------------
	void heroMeleeAttack (Hero attacker, Unit defender) {
		if (attacker.getFigures() < 1) return;

		// Skip wizards in melee
		if (attacker.isWizard())
			return;

		// Compute figures in attack
		double atkWidth = attacker.getFiles() * attacker.getFigWidth();
		double defWidth = defender.getFiles() * defender.getFigWidth();
		int figsAtk = (atkWidth <= defWidth ? attacker.getFiles() :
			(int) Math.ceil(defWidth / attacker.getFigWidth()));

		// Check dragon breath attack
		if (useSpecialAbilities && attacker.isDragon()
				&& attacker.getCharges() > 0)
			dragonBreathAttack(attacker, defender, figsAtk);

		// Roll attack dice
		int numHits = 0;
		for (int i = 0; i < figsAtk; i++) {
			if (d6() >= attacker.getAtk()) {
				numHits += Math.min(attacker.getDam(), defender.getHD());
			}
		}

		// Apply damage
		applyDamage(attacker, defender, false, numHits);
	}

	//-----------------------------------------------------------------
	//  Magic area attack
	//    Adjudicates one hit on one basic figure
	//    (If target is attached hero, identify that as target)
	//    Returns damage taken
	//-----------------------------------------------------------------
	int magicAreaAttack (Unit target, int damage) {
		if (!(target instanceof Hero)) {
			int save = (int) (Math.random() * 3) + 1; // d3
			int damageTaken = Math.min(damage - save, target.getHD());
			return (damageTaken > 0 ? damageTaken : 0);
		}
		else {
			int save = d6();
			return (save < damage/2 ? 1 : 0);
		}
	}

	//-----------------------------------------------------------------
	//  Dragon breath attack
	//-----------------------------------------------------------------
	void dragonBreathAttack (Hero dragon, Unit target, int figsAtk) {
		assert(dragon.getCharges() > 0);
		if (dragon.getFigures() < 1) return;
		reportDetail("* DRAGON BREATH ATTACK *");
		dragon.decrementCharges();

		// Get damage by type (adults) & check immunities
		int damage = 0;
		if (dragon.nameContains("Blue")) {
			if (target.nameContains("Giant, Storm")) return;
			else damage = 10;
		}
		else if (dragon.nameContains("Red")) {
			if (target.nameContains("Giant, Fire")) return;
			else damage = 11;
		}
		else if (dragon.nameContains("Gold"))
			damage = 12;
		
		// Assess hits
		int hits = 0;
		for (int i = 0; i < figsAtk; i++) {
			hits += magicAreaAttack(target, damage);
		}
		applyDamage(dragon, target, true, hits);
	}

	//-----------------------------------------------------------------
	//  Wizard missile hit assessment (fireball, lightning bolt)
	//    Assumes variation of +/-1 inch
	//-----------------------------------------------------------------
	boolean missileHit (Unit target) {
		double targetDepth = target.getFigLength() * target.getRanks();
		if (targetDepth <= 0.0)
			return false;
		else if (targetDepth <= 1.0) 
			return (d6() <= 2);	// 2-in-6
		else if (targetDepth <= 2.0)
			return (d6() <= 4);  // 4-in-6
		else
			return true;
	}

	//-----------------------------------------------------------------
	//  Wizard spell attack (assumes motionless; full fire)
	//-----------------------------------------------------------------
	void wizardSpellAttack (Hero wizard, Unit target) {
		assert(wizard.isWizard());
		if (wizard.getFigures() < 1) return;

		// Assume each greater-spell charge used for "Death Spell"
		if (wizard.getCharges() > 0 && distance <= 12
				&& target.getHD() <= 8 && !(target instanceof Hero)) {
			reportDetail("* WIZARD DEATH SPELLS *");
			int hits = 0;
			for (int i = 0; i < wizard.getFigures(); i++)
				hits += 4 + wizard.getRanks() / 3;
			applyDamage(wizard, target, true, hits);		
			wizard.decrementCharges();
		}

		// Otherwise shoot two wand-based "Fireballs"
		else if (distance <= 24) {
			reportDetail("* WIZARD FIREBALLS *");
			int hits = 0;
			for (int i = 0; i < wizard.getFigures() * 2; i++) {
				if (missileHit(target))
					hits += magicAreaAttack(target, 6);
			}
			if (useSpecialAbilities && target.nameContains("Giant, Fire"))
				hits = 0;
			applyDamage(wizard, target, true, hits);		
		}
	}

	//-----------------------------------------------------------------
	//  Is the defender immune to this attack?
	//-----------------------------------------------------------------
	boolean isAttackImmune (Unit atk, Unit def) {
		assert(!(atk instanceof Hero));
		if (useSpecialAbilities) {
			if (atk.getHD() < 6 && def.nameStarts("Elemental"))
				return true; // AD&D rule vs. +2 to hit
		}
		return false;
	}

	//-----------------------------------------------------------------
	//  Does this unit get the pike special benefit?
	//-----------------------------------------------------------------
	boolean hasPikeBenefit (Unit unit) {

		boolean hasBenefit 
			= unit.hasPikes()
			&& !pikeFlanking
			&& !unit.isBeaten()
			&& terrain == Terrain.Open;

		return hasBenefit;
	}

	//-----------------------------------------------------------------
	//  Is the special pike interrupt attack happening now?
	//-----------------------------------------------------------------
	boolean onPikeInterrupt (Unit unit) {
		return hasPikeBenefit(unit)
			&& !pikeInterrupted && distance == 0;
	}
	
	//-----------------------------------------------------------------
	//  Basic attack roll
	//-----------------------------------------------------------------
	boolean rollToHit (int AH, int HD, int mods) {
		int roll = d6();
		int target = AH - HD/3 - mods;
		return (roll >= target);
	}

	//-----------------------------------------------------------------
	//  Check morale (sets field if failed)
	//-----------------------------------------------------------------
	void checkMorale (Unit unit) {
		if (unit.getFigures() == 0 
			|| unit.getFigsLostInTurn() == 0) return;
		if (isMoraleImmune(unit)) return;

		// Get terms
		int roll = d6()+d6();
		int hitDice = unit.getHD();
		int rateOfLoss = unit.getFigures() / unit.getFigsLostInTurn();
		int miscBonus = miscMoraleBonus(unit);

		// Total, report, and assessment
		int check = roll + hitDice + rateOfLoss + miscBonus;
		reportDetail("Morale check (" + unit + "): " + check);
		if (check < 10) {
			unit.setRouted(true);
			reportDetail(unit + " are *ROUTED*");
		}
	}

	//-----------------------------------------------------------------
	//  Miscellaneous morale bonuses
	//-----------------------------------------------------------------
	int miscMoraleBonus (Unit unit) {
		int bonus = 0;

		// Light weakness (orc, goblin)
		if (useSpecialAbilities && weather == Weather.Sunny
				&& (unit.nameStarts("Orc") || unit.nameStarts("Goblin")))
			bonus -= 1;

		// Optional stuff
		if (useOptionalMoraleMods) {

			// Alignment
			if (unit.getAlignment() == Unit.Alignment.Lawful) 
				bonus += 1;
			else if (unit.getAlignment() == Unit.Alignment.Chaotic) 
				bonus -= 1;

			// Cohesion is skipped here (only 1 unit in question)

			// Extra Ranks
			if (unit.getRanks() > 1)
				bonus += Math.min(unit.getRanks()-1, 3);

			// Leadership
			if (unit.hasHero() && unit.getHero().getFigures() > 0)
				bonus += 1;
		}
		
		return bonus;		
	}
	
	//-----------------------------------------------------------------
	//  Is this unit immune to morale (fearless)?
	//-----------------------------------------------------------------
	boolean isMoraleImmune (Unit unit) {
		if (unit instanceof Hero) return true;	
		if (useSpecialAbilities && unit.nameStarts("Troll")) return true;
		return false;
	}

	//-----------------------------------------------------------------
	//  Randomize terrain
	//    Assume just one terrain type across entire field
	//    Percents match coverage of about 37% middle of table
	//-----------------------------------------------------------------
	void randomizeTerrain () {
		int roll = (int) (Math.random() * 100);
		roll = (int) (roll/terrainMultiplier);
		if (roll < 2) terrain = Terrain.Gulley;
		else if (roll < 5) terrain = Terrain.Rough;
		else if (roll < 15) terrain = Terrain.Hill;
		else if (roll < 29) terrain = Terrain.Woods;
		else if (roll < 34) terrain = Terrain.Marsh;
		else if (roll < 37) terrain = Terrain.Stream;
		else terrain = Terrain.Open;		
		// Note: Pond type is not used (impassable)
	}

	//-----------------------------------------------------------------
	//  Randomize weather
	//-----------------------------------------------------------------
	void randomizeWeather () {
		int roll = d6()+d6();
		if (roll <= 7) weather = Weather.Sunny;
		else if (roll <= 9) weather = Weather.Cloudy;
		else if (roll <= 11) weather = Weather.Rainy;
		else weather = Weather.Stormy;
	}

	//-----------------------------------------------------------------
	//  Roll one 6-sided die
	//-----------------------------------------------------------------
	int d6 () {
		return (int) (Math.random() * 6) + 1;
	}

	//-----------------------------------------------------------------
	//  Main method
	//-----------------------------------------------------------------
	public static void main (String[] args) {
		BookOfWar sim = new BookOfWar();
		if (gameDetail)
			sim.gameDetail();
		else 
			sim.assessmentTable();
		Toolkit.getDefaultToolkit().beep();			
	}
}


/*
=====================================================================
LICENSING INFORMATION

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA

The author may be contacted by email at: delta@superdan.net
=====================================================================
*/
