//********************************************************************
//  Hero.java
//    Represents a hero unit (posibly wizard).
//  Copyright (c) 2010-11 Daniel R. Collins. All rights reserved.
//********************************************************************

public class Hero extends Unit {
	private int atk, dam;      // Attack to-hit; Damage amount
	private int wizardRank;    // Wizard rank = level - 10
	private boolean dragon;
	private int specialCharges; 

	//-----------------------------------------------------------------
	//  Constructor (melee-types)
	//-----------------------------------------------------------------
	public Hero(String name, Type type, int cost, 
			int MV, int AH, int HD, int atk, int dam) {
		super(name, type, cost, MV, AH, HD);
		this.atk = atk;
		this.dam = dam;					
		this.wizardRank = 0;
		dragon = name.contains("Dragon");
	}

	//-----------------------------------------------------------------
	//  Constructor (wizards)
	//-----------------------------------------------------------------
	public Hero(String name, Type type, int cost, 
			int MV, int AH, int HD, int atk, int dam, int wizardRank) {
		this(name, type, cost, MV, AH, HD, atk, dam);
		this.wizardRank = wizardRank;
	}

	//-----------------------------------------------------------------
	//  Accessors
	//-----------------------------------------------------------------
	public int getAtk() { return atk; }
	public int getDam() { return dam; }
	public int getCharges() { return specialCharges; }
	public int getWizardRank() { return wizardRank; }
	public boolean isWizard() { return wizardRank > 0; }
	public boolean isDragon() { return dragon; }
	
	//-----------------------------------------------------------------
	//  Refresh special ability charges
	//-----------------------------------------------------------------
	public void refreshCharges() { 
		if (isWizard()) // 6th-level spells
			specialCharges = wizardRank - 1;			
		else if (isDragon())
			specialCharges = 3;
		else
			specialCharges = 0;	
	}
	
	//-----------------------------------------------------------------
	//  Decrement special ability charges
	//-----------------------------------------------------------------
	public void decrementCharges() {
		specialCharges--;	
	}
}
